Lock functions for SMS/QDOS                                      January 1996

NAME
        GetLock:        Blocking exclusive lock access (locks 0-7)
        GetLock0:       Blocking exclusive lock access (lock 0)
        FreeLock:       Release exclusive lock (locks 0-7)
        FreeLock0:      Release exclusive lock (lock 0)
        InitLock:       Initialise a lock structure
        TestLock:       Non-blocking exclusive lock access (locks 0-7)
        TestLock0:      Non-blocking exclusive lock access (lock 0)

SYNOPSIS:

        #include <lock.h>

        void GetLock(int lockNo, LOCK lock);
        void GetLock0(LOCK lock);
        void FreeLock(int lockNo, LOCK lock);
        void FreeLock0(LOCK lock);
        void InitLock(LOCK lock);
        int  TestLock(int lockNo, LOCK lock);
        int  TestLock0(LOCK lock);

DESCRIPTION:

        The lock functions provide a means of inter-job locking
        (semaphores). Each LOCK structure may hold up to eight locks,
        (denoted lock 0 through lock 7) with no limit on the number of
        LOCK structures available to jobs. LOCKs may be static to a
        job to provide locking within a multi-threaded job, or in a
        global data area (Thing ?) to provide locking between
        independent jobs.

        The '0' varients of the functions provide access to lock 0
        only, with slightly lower overhead.

        The InitLock function initialises a LOCK structure. This
        should be done before a lock is used. This should only be done
        once (by the master job). Failure to heed this may result in
        unpredictable program behaviour (i.e. it will crash).

        If the LOCK structure is declared such that it is initialised
        to zero (i.e. global or static), the it need not be
        initialised with InitLock().

        The TestLock function attempts to gain access to a lock. If
        the lock is granted, the function returns 0. If the function
        returns -1, then the lock is in use by another job.

        The GetLock function gains access to a lock. The function will
        block until access to the lock is granted.

        The FreeLock function releases a previously obtained lock. It
        is unadvisable to release a lock you do not own.

        A macro is provided to test the lock status.

        LOCK_QUEUE(lock) returns non-zero if any of the eight locks
        controlled by the LOCK 'lock' have jobs waiting for them. It
        returns 0 if no jobs are waiting.

USAGE:
        Assemble lock.s with as68 and link lock.o with your program.
        Include "lock.h" in your program.

EXAMPLE:

    #include <stdio.h>
    #include <stdlib.h>
    #include <lock.h>
    #include <qdos.h>
    #include <unistd.h>

    /*
     * Lock demo (c) jh '96 - No rights reserved
     * build as: qcc -I. -O -o testl testl.c lock.s jcall.s
     * for Linux/XTC68, SMS/QDOS as appropriate
     */

    void (*_SigStart)(void) = NULL;
    static LOCK lck;

    void thread(int n)
    {
        int l1;
        GetLock0(lck);
        l1 = TestLock(1, lck);
        printf("Thread %d got the lock 0", n);
        if(l1 == 0)
        {
            puts(", also got lock 1");
        }
        else
        {
            puts(", exiting");
        }

        FreeLock0(lck);

        if(l1 == 0)
        {
            mt_susjb(-1, 150, NULL);
            printf("Thread %d giving up lock 1\n", n);
            FreeLock(1, lck);
        }
    }

    extern jcall (void (*)( ), char *, int, int, int,...);

    int main(void)
    {
        InitLock(lck);

        GetLock0(lck);
        puts("Parent, locked 0");

        /*
         * Now start up two indepentent threads that can fight over 
         * lock 0 .. as its a LIFO, my guess is Thread2 will win
         */

        jcall(thread, "Thread1", 1, 1024, 1, 1);
        jcall(thread, "Thread2", 1, 1024, 1, 2);

        /* Wait until a thread asks for the lock */

        while(LOCK_QUEUE(lck) == 0)
        {
            mt_susjb(-1, 2, NULL);
        }
        FreeLock0(lck);

        /* Now the threads will unblock */
        /* We block to make sure a thread gets to lock 1 first */

        while(LOCK_QUEUE(lck))
        {
            mt_susjb(-1, 2, NULL);
        }

        /* The lock queues are empty, wait until the last lock is freed */

        GetLock(1, lck);
        fputs("main is exiting : ", stdout);
        return 0;
    }



BUGS:
        The functions cannot detect or prevent deadlock.

CREDITS: The example program uses the freely distributable jcall
        function by Tony Tebby (from the libqptr package).

AUTHOR:
        Jonathan Hudson. No rights reserved. 12/01/96
